# Copyright 2025 Bosutech XXI S.L.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field

from nucliadb_models.common import QuestionAnswers
from nucliadb_models.extracted import FieldMetadata
from nucliadb_models.text import FieldText


class AgentType(Enum):
    graph = "graph"
    label = "label"
    ask = "ask"
    qa = "qa"
    extract = "extract"
    prompt_guard = "prompt_guard"
    llama_guard = "llama_guard"


class AgentsFilter(BaseModel):
    type: AgentType
    task_names: list[str] = Field(
        default_factory=list,
        description="list of task names. If None or empty, all tasks for that operation are applied.",
    )


class ResourceAgentsRequest(BaseModel):
    filters: Optional[list[AgentsFilter]] = Field(
        default=None,
        description="Filters to apply to the agents. If None, all curently configured agents are applied.",
    )


class NewTextField(BaseModel):
    text_field: FieldText
    destination: str


class AppliedDataAugmentation(BaseModel):
    qas: Optional[QuestionAnswers] = Field(
        default=None,
        description="Question and answers generated by the Question Answers agent",
    )
    new_text_fields: list[NewTextField] = Field(
        default_factory=list,
        description="New text fields. Only generated by the Generator agent as of now.",
    )
    changed: bool = Field(
        default=True,
        description="Indicates if the FieldMetadata was changed by the agents",
    )


class AugmentedField(BaseModel):
    metadata: FieldMetadata = Field(
        ...,
        title="The updated metadata of the field",
    )
    applied_data_augmentation: AppliedDataAugmentation = Field(
        ..., title="The results of the Applied Data Augmentation"
    )
    input_nuclia_tokens: float = Field(
        ..., title="The number of input Nuclia tokens consumed for the field"
    )
    output_nuclia_tokens: float = Field(
        ..., title="The number of output Nuclia tokens consumed for the field"
    )
    time: float = Field(..., title="The time taken to execute the Data Augmentation agents to the field")


class ResourceAgentsResponse(BaseModel):
    results: dict[str, AugmentedField] = Field(
        ...,
        title="Pairs of augmented FieldMetadata and Data Augmentation results by field id",
    )
